import argparse
import sys
from .shared import glob_paths_files, eprint, print_utf8
from . import chunks, parse_size

def print_hex(data, seek):
    for i, chunk in enumerate(chunks(data, 16)):
        print_utf8("{:08x}".format(i * 16 + seek), end=' ')
        chars = ["{:02x}".format(c) for c in chunk] + ["  " for _ in range(len(chunk), 16)]
        print_utf8(" ".join(chars), end='  ')
        text = chunk.decode('utf-8', errors='replace').replace('\n', '.').replace('\r','.')
        print_utf8(text)

def seek_and_read(f, seek, len_):
    f.seek(seek)
    size = len_ if len_ > -1 else None
    data = f.read(size)
    return data

def main():
    parser = argparse.ArgumentParser(description='prints file as hex')
    parser.add_argument('path', nargs='*')
    parser.add_argument('-s','--seek', help='start at offset')
    parser.add_argument('-l','--len', help='number of bytes to print')

    args = parser.parse_args()

    if args.seek is None:
        seek = 0
    else:
        seek = parse_size(args.seek)

    if args.len is None:
        len_ = -1
    else:
        len_ = parse_size(args.len)

    if len(args.path) == 0:
        data = seek_and_read(sys.stdin.buffer, seek, len_)
        print_hex(data, seek)
    else:
        for path in glob_paths_files(args.path):
            try:
                with open(path, 'rb') as f:
                    data = seek_and_read(f, seek, len_)
                print_hex(data, seek)
            except Exception as e:
                eprint(e)

if __name__ == "__main__":
    main()


import re
import os
import argparse

from .shared import glob_paths, glob_paths_dirs
import datetime
import json
import fnmatch

def median(vs):
    if len(vs) % 2:
        return vs[len(vs) // 2]
    else:
        return (vs[len(vs) // 2 - 1] + vs[len(vs) // 2]) / 2

def average(vs):
    return sum(vs) / len(vs)

def parse_time_arg(v):
    m = re.match('([0-9.e-]+)(d|h|m|s|)', v)
    v = float(m.group(1))
    k = {"d": 3600 * 24, "h": 3600, "m": 60, "s": 1, "": 1}[m.group(2)]
    return k * v

def include_exclude(include, exclude, name):
    if include is None and exclude is None:
        return True
    if include is None:
        ok = True
    else:
        ok = False
        for pat in include:
            if fnmatch.fnmatch(name, pat):
                ok = True
    if not ok:
        return False
    if exclude is None:
        return True
    else:
        for pat in exclude:
            if fnmatch.fnmatch(name, pat):
                return False
    return True

def main():
    parser = argparse.ArgumentParser(description='prints mtime statistics')
    parser.add_argument('path', nargs='*')

    parser.add_argument('-r', '--recent', help='print recently changed files (max(mtime) - mtime <= N)')
    parser.add_argument('-o', '--old', help='print long ago changed files (mtime - min(mtime) <= N)')
    parser.add_argument('-a', '--average', help='print files changed within N seconds around average (abs(average(mtime) - mtime) <= N)')
    parser.add_argument('-i', '--include', nargs='+', help='include files glob')
    parser.add_argument('-e', '--exclude', nargs='+', help='include files glob')

    args = parser.parse_args()

    #print(args)

    stat = []

    if len(args.path) == 0:
        paths = ["."]
    else:
        paths = glob_paths_dirs(args.path)
    mtime = []
    for path in paths:
        for root, dirs, files in os.walk(path):
            for f in files + dirs:
                if not include_exclude(args.include, args.exclude, f):
                    continue
                p = os.path.join(root, f)
                t = os.path.getmtime(p)
                stat.append({"t":t, "p": p})
                mtime.append(t)

    if len(mtime) == 0:
        print("No files")
        exit(1)

    mtime.sort()

    names = ['min', 'max', 'average', 'median']

    min_mtime = min(mtime)
    max_mtime = max(mtime)
    avg_mtime = average(mtime)

    values = [min_mtime, max_mtime, avg_mtime, median(mtime)]

    if args.recent or args.old or args.average:
        
        if args.recent:
            arg_recent = parse_time_arg(args.recent)
        else:
            arg_recent = None

        if args.old:
            arg_old = parse_time_arg(args.old)
        else:
            arg_old = None
        
        if args.average:
            arg_average = parse_time_arg(args.average)
        else:
            arg_average = None

        #print("arg_recent {}, arg_old {}, arg_average {}".format(arg_recent, arg_old, arg_average))

        for path in paths:
            for root, dirs, files in os.walk(path):
                for f in files + dirs:

                    if not include_exclude(args.include, args.exclude, f):
                        continue

                    p = os.path.join(root, f)
                    t = os.path.getmtime(p)

                    do_print = False
                    if arg_recent is not None:
                        if max_mtime - t <= arg_recent:
                            do_print = True
                    if arg_old is not None:
                        if t - min_mtime <= arg_old:
                            do_print = True
                    if arg_average is not None:
                        if abs(t - avg_mtime) <= arg_average:
                            do_print = True
                    
                    if do_print:
                        if os.path.isabs(path):
                            p_ = p
                        else:
                            p_ = os.path.relpath(p, os.getcwd())
                        print("{} {}".format(datetime.datetime.fromtimestamp(t), p_))

    print("{:8} files".format(len(mtime)))
    for n,v in zip(names, values):
        print("{:8} {}".format(n, datetime.datetime.fromtimestamp(v)))

if __name__ == "__main__":
    main()


from ast import parse
import os
import argparse
import glob
import subprocess
import shutil

def find_explorer():
    # todo other platforms
    explorer = shutil.which('explorer')
    if explorer is not None:
        return explorer

    root = os.environ['SystemRoot']
    path = os.path.join(root, "explorer.exe")
    if os.path.exists(path):
        return path
    path = os.path.join(root, "system32", "explorer.exe")
    if os.path.exists(path):
        return path

def main():
    parser = argparse.ArgumentParser(description='opens file in associated application or directory in explorer')
    parser.add_argument('path', nargs='+', help='paths to start')
    parser.add_argument('--show', action='store_true')
    args = parser.parse_args()

    paths = []

    for path in args.path:
        if glob.has_magic(path):
            for path_ in glob.glob(path):
                paths.append(path_)
        else:
            paths.append(path)

    if args.show:
        explorer = find_explorer()
    
    for path in paths:
        if args.show:
            subprocess.run([explorer, "/select,", path])
        else:
            os.startfile(path)

if __name__ == "__main__":
    main()


import argparse
from random import uniform
from collections import defaultdict
from .shared import glob_paths_files, read_lines, print_lines

def main():
    parser = argparse.ArgumentParser(description='prints unique or nonunique lines from sorted array of lines')
    parser.add_argument('--count', '-c', action='store_true')
    parser.add_argument('--repeated','-d',action='store_true')
    parser.add_argument('--unique','-u',action='store_true')
    parser.add_argument('path', nargs="*")

    args = parser.parse_args()
    paths = glob_paths_files(args.path)
    lines = read_lines(paths)

    stat = defaultdict(lambda: 0)
    for line in lines:
        stat[line] += 1
    
    if args.repeated:
        lines_ = [line for line in lines if stat[line] > 1]
    elif args.unique:
        lines_ = [line for line in lines if stat[line] == 1]
    else:
        lines_ = lines

    res = [lines_[0]]
    for line in lines_:
        if res[-1] != line:
            res.append(line)

    if args.count:
        formatter = lambda line: "{:5d} {}".format(stat[line], line)
    else:
        formatter = lambda line: line

    lines_ = [formatter(line) for line in res]
    print_lines(lines_)

if __name__ == "__main__":
    main()


import os
import argparse
from .shared import glob_paths, print_utf8
from . import read_stdin_text
import datetime

def print_mtime(path):
    try:
        print_utf8("{} {}".format(datetime.datetime.fromtimestamp(os.path.getmtime(path)), path))
    except Exception as e:
        pass

def main():
    parser = argparse.ArgumentParser(description='prints mtime of file')
    parser.add_argument('path', nargs='*')
    args = parser.parse_args()
    
    if len(args.path) == 0:
        for line in read_stdin_text().split('\n'):
            print_mtime(line.rstrip())
    else:
        paths = glob_paths(args.path)
        for path in paths:
            print_mtime(path)

if __name__ == "__main__":
    main()


import argparse
from .shared import glob_paths, print_utf8
import glob
import os

def main():
    parser = argparse.ArgumentParser(description='lists directory')
    parser.add_argument('path', nargs='*')
    args = parser.parse_args()

    if len(args.path) == 0:
        paths = ['.']
    else:
        paths = args.path

    for path in paths:
        if glob.has_magic(path):
            for f in glob.glob(path):
                print_utf8(f)
        else:
            if len(paths) > 1:
                print_utf8(path + ":")
            for root, dirs, files in os.walk(path):
                for d in dirs:
                    print_utf8(d + "/")
                for f in files:
                    print_utf8(f)
                while len(dirs) > 0:
                    dirs.pop()
    

if __name__ == "__main__":
    main()


import os
import argparse
import zipfile
from .shared import print_utf8, glob_paths

def main():
    parser = argparse.ArgumentParser(description='appends, extracts and list contents of zip archive')
    parser.add_argument('command', choices=['a','x','l'], help='add extract list')
    parser.add_argument('-o', '--output', help='output directory')
    parser.add_argument('zip')
    parser.add_argument('sources', nargs='*') # todo wildcards
    args = parser.parse_args()

    dst_base = args.output if args.output is not None else os.getcwd()

    if args.command == 'a':
        paths = glob_paths(args.sources)
        with zipfile.ZipFile(args.zip, 'a') as zf:
            for path in paths:
                #base = os.path.dirname(path)
                base = os.getcwd()
                for root, dirs, files in os.walk(path):
                    for f in files:
                        p = os.path.join(root, f)
                        relpath = os.path.relpath(p, base)
                        zf.write(p, relpath)

    if args.command == 'x':
        with zipfile.ZipFile(args.zip) as zf:
            for name in zf.namelist():
                name_ = name.replace('/','\\')
                if len(args.sources) == 0 or name in args.sources or name_ in args.sources:
                    #dest = os.path.join(dst_base, name)
                    print("extract {} to {}".format(name, dst_base))
                    zf.extract(name, dst_base)
                    
    if args.command == 'l':
        with zipfile.ZipFile(args.zip) as zf:
            for name in zf.namelist():
                print_utf8(name)

if __name__ == "__main__":
    main()


import sys
import re
from .shared import glob_paths, glob_paths_files, read_bytes, print_bytes
import argparse
import hashlib

def read_stdin_text():
    data = sys.stdin.buffer.read()
    try:
        text = data.decode('utf-8')
    except UnicodeDecodeError:
        text = data.decode(sys.stdin.encoding)
    return text

def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def parse_size(arg):
    m = re.match('^([-+]?[0-9.e]+)+([cwbkmg]?)$', arg, re.IGNORECASE)
    if m is None:
        return None
    size = float(m.group(1))
    prefix = m.group(2)
    return int(size * {'k': 1024, 'm': 1024 * 1024, 'g': 1024 * 1024 * 1024, 'c': 1, 'b': 512, '': 1}[prefix.lower()])

import argparse

(
    TO_UNIX,
    TO_DOS
) = range(2)

def convertlineterm(to_what):

    description = {
        TO_UNIX: 'converts line endings from dos to unix (\\r\\n -> \\n)',
        TO_DOS: 'converts line endings from unix to dos (\\n -> \\r\\n)'
    }[to_what]

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('path', nargs='*')
    args = parser.parse_args()
    paths = glob_paths_files(args.path)

    stdin_mode = len(args.path) == 0

    if to_what == TO_UNIX:
        subj = b'\r\n'
        repl = b'\n'
    else:
        subj = b'\n'
        repl = b'\r\n'

    if stdin_mode:
        bytes_ = read_bytes(None)
        print_bytes(bytes_.replace(subj, repl))
    else:
        for path in paths:
            bytes_ = read_bytes(path)
            with open(path, 'wb') as f:
                f.write(bytes_.replace(subj, repl))


def files_hash(paths, from_stdin, alg):
    if from_stdin:
        hash = hashlib.new(alg)
        hash.update(sys.stdin.buffer.read())
        print(" ".join([hash.hexdigest(), "-"]))
    else:
        for path in paths:
            with open(path, 'rb') as f:
                hash = hashlib.new(alg)
                hash.update(f.read())
                print(" ".join([hash.hexdigest(), path]))

def files_hash_main(alg):
    parser = argparse.ArgumentParser(description='prints {} hashsum of file'.format(alg))
    parser.add_argument('path', nargs='*')
    args = parser.parse_args()
    from_stdin = len(args.path) == 0
    paths = glob_paths_files(args.path)
    files_hash(paths, from_stdin, alg)
    



import argparse
import os
import datetime
import sys
import re
import dateutil.parser
import sys
from dataclasses import dataclass
from itertools import count
import subprocess
import shutil
from .shared import eprint, run, print_utf8
import fnmatch
from . import parse_size

class Tok:
    (
        und,
        op_par,
        cl_par,
        not_,
        and_,
        or_,
        mmin,
        iname,
        name,
        type,
        f,
        d,
        newer,
        newermt,
        newerct,
        arg,
        path,
        ipath,
        mtime,
        ctime,
        size,
        exec,
        delete,
        semicolon,
        pathbind,
        icont,
        cont,
        bcont,
        maxdepth,
    ) = range(29)
    
m = {
    "(": Tok.op_par,
    ")": Tok.cl_par,
    "!": Tok.not_,
    "-not": Tok.not_,
    "-a": Tok.and_,
    "-and": Tok.and_,
    "-o": Tok.or_,
    "-or": Tok.or_,
    "-mmin": Tok.mmin,
    "-iname": Tok.iname,
    "-name": Tok.name,
    "-type": Tok.type,
    "f": Tok.f,
    "d": Tok.d,
    "-newer": Tok.newer,
    "-newermt": Tok.newermt,
    "-newerct": Tok.newerct,
    "-mtime": Tok.mtime,
    "-ctime": Tok.ctime,
    "-size": Tok.size,
    "-exec": Tok.exec,
    "-delete": Tok.delete,
    ";": Tok.semicolon,
    "{}": Tok.pathbind,
    "-icont": Tok.icont,
    "-cont": Tok.cont,
    "-bcont": Tok.bcont,
    "-path": Tok.path,
    "-ipath": Tok.ipath,
    "-maxdepth": Tok.maxdepth,
}

inv_m = {v:k for k,v in m.items()}

@dataclass
class T:
    type: int
    cont: str

tok_pred = [Tok.mmin, Tok.iname, Tok.name, Tok.type, Tok.newer, 
    Tok.newerct, Tok.newermt, Tok.mtime, Tok.ctime,  Tok.size, Tok.cont, Tok.icont, Tok.bcont, Tok.path, Tok.ipath]

class Action:
    def __init__(self, tokens):
        self._tokens = tokens

    def exec(self, root, name, path, is_dir):
        expr = [path if t.cont == '{}' else t.cont for t in self._tokens]
        run(expr)

class ActionDelete:
    def exec(self, root, name, path, is_dir):
        if is_dir:
            eprint("Removing directory {}".format(path))
            shutil.rmtree(path)
        else:
            eprint("Removing file {}".format(path))
            os.remove(path)


class ActionPrint:
    def exec(self, root, name, path, is_dir):
        if os.path.isabs(root):
            print_utf8(path)
        else:
            print_utf8(os.path.relpath(path, os.getcwd()))

def index_of_token(tokens, type):
    for i, tok in enumerate(tokens):
        if tok.type == type:
            return i

def parse_args(args = None):
    if args is None:
        args = sys.argv[1:]
    tokens = [T(m[t], t) if t in m else T(Tok.und, t) for t in args]
    for i, tok in enumerate(tokens):
        if tok.type in tok_pred:
            token = tokens[i+1]
            tokens[i+1] = T(Tok.arg, token.cont)

    maxdepth = 0
    ix_maxdepth = index_of_token(tokens, Tok.maxdepth)
    if ix_maxdepth is not None:
        tokens.pop(ix_maxdepth)
        token = tokens.pop(ix_maxdepth)
        maxdepth = int(token.cont)

    action = ActionPrint()

    ix_delete = index_of_token(tokens, Tok.delete)
    if ix_delete is not None:
        tokens.pop(ix_delete)
        action = ActionDelete()
    
    ix_exec = index_of_token(tokens, Tok.exec)
    ix_semicolon = index_of_token(tokens, Tok.semicolon)

    if ix_exec is not None:
        if ix_semicolon is None:
            raise ValueError("Invalid exec expression: semicolon not found")
        exec_tokens = tokens[ix_exec+1:ix_semicolon]
        action = Action(exec_tokens)
        tokens = tokens[:ix_exec] + tokens[ix_semicolon+1:]
        #print(tokens)

    i = len(tokens)-1
    while i > -1 and tokens[i].type == Tok.und:
        i -= 1
    tokens, paths1 = tokens[:i+1], tokens[i+1:]
    i = 0
    while i < len(tokens) and tokens[i].type == Tok.und:
        i += 1
    paths2 = []
    if i > 0:
        paths2, tokens = tokens[:i], tokens[i:]

    expr = tokens
    paths = [e.cont for e in paths1 + paths2]

    #print('expr, paths', expr, paths)

    return expr, paths, action, maxdepth


class Cache:

    # todo optimize cache for perfomance and memory (do dictionary lookups become costly for huge dictionaries?)

    def __init__(self):
        self._mtime = dict()
        self._ctime = dict()
        self._size = dict()
        self._parse_size = dict()
        self._now = datetime.datetime.now()
        self._parse_date = dict()

    def mtime(self, path):
        if path not in self._mtime:
            self._mtime[path] = None
            try:
                self._mtime[path] = datetime.datetime.fromtimestamp(os.path.getmtime(path))
            except Exception as e:
                eprint(e)
        return self._mtime[path]

    def ctime(self, path):
        if path not in self._ctime:
            self._ctime[path] = None
            try:
                self._ctime[path] = datetime.datetime.fromtimestamp(os.path.getctime(path))
            except Exception as e:
                eprint(e)
        return self._ctime[path]

    def size(self, path):
        if path not in self._size:
            self._size[path] = None
            try:
                self._size[path] = os.path.getsize(path)
            except Exception as e:
                eprint(e)
        return self._size[path]

    def parse_date(self, date):
        if date not in self._parse_date:
            self._parse_date[date] = None
            try:
                self._parse_date[date] = dateutil.parser.parse(date)
            except Exception as e:
                eprint("Invalid datetime format {}".format(date))
        return self._parse_date[date]

    def parse_size(self, arg):
        if arg not in self._parse_size:
            self._parse_size[arg] = None
            parsed = parse_size(arg)
            if parsed is None:
                eprint("Invalid size format {}".format(arg))
            else:
                self._parse_size[arg] = parsed
        return self._parse_size[arg]

    def now(self):
        return self._now

def pred_mmin(name, path, is_dir, arg, cache):
    mtime = cache.mtime(path)
    if mtime is None:
        return None
    total_min = (cache.now() - mtime).total_seconds() / 60
    arg = float(arg)
    if arg < 0:
        return total_min < abs(arg)
    return total_min > arg

def pred_iname(name, path, is_dir, arg, cache):
    return fnmatch.fnmatch(name, arg)

def pred_ipath(name, path, is_dir, arg, cache):
    return fnmatch.fnmatch(path, arg)

def pred_path(name, path, is_dir, arg, cache):
    return fnmatch.fnmatchcase(path, arg)

def pred_type(name, path, is_dir, arg, cache):
    # todo validate type arg
    return is_dir == (arg == 'd')

def greater(d1, d2):
    if None in [d1, d2]:
        return None
    return d1 > d2

def pred_newer(name, path, is_dir, arg, cache):
    #print([cache.mtime(path), cache.mtime(arg), cache.mtime(path) > cache.mtime(arg), arg])
    return greater(cache.mtime(path), cache.mtime(arg))
    
def pred_newermt(name, path, is_dir, arg, cache):
    return greater(cache.mtime(path), cache.parsedate(arg))

def pred_newerct(name, path, is_dir, arg, cache):
    return greater(cache.ctime(path), cache.parsedate(arg))

def pred_xtime(arg, cache, xtime):
    if xtime is None:
        return None
    total_days = (cache.now() - xtime).total_seconds() / 60 / 60 / 24
    arg = float(arg)
    if arg < 0:
        return total_days < abs(arg)
    return total_days > arg

def pred_mtime(name, path, is_dir, arg, cache):
    return pred_xtime(arg, cache, cache.mtime(path))

def pred_ctime(name, path, is_dir, arg, cache):
    return pred_xtime(arg, cache, cache.ctime(path))

def pred_size(name, path, is_dir, arg, cache):
    if is_dir:
        return None
    size_arg = cache.parse_size(arg)
    size_path = cache.size(path)
    if None in [size_arg, size_path]:
        return None
    if size_arg < 0:
        return size_path < abs(size_arg)
    return size_path > size_arg

def pred_xcont(name, path, is_dir, arg, cache, flags, bin):
    if is_dir:
        return None
    try:
        with open(path, 'rb') as f:
            data = f.read()
        if bin:
            arg = arg.encode("utf-8").decode('unicode_escape').encode("utf-8")
        else:
            arg = arg.encode("utf-8")
        return re.search(arg, data, flags) is not None

    except Exception as e:
        eprint(e)
    return None

def pred_icont(name, path, is_dir, arg, cache):
    return pred_xcont(name, path, is_dir, arg, cache, re.IGNORECASE, False)

def pred_cont(name, path, is_dir, arg, cache):
    return pred_xcont(name, path, is_dir, arg, cache, 0, False)

def pred_bcont(name, path, is_dir, arg, cache):
    return pred_xcont(name, path, is_dir, arg, cache, 0, True)

def max_level(tree):
    level = -1
    for e in tree:
        if isinstance(e, tuple):
            level = max(level, e[1])
    return level

def index_op_par(tree, level):
    for i, e in enumerate(tree):
        if isinstance(e, tuple):
            if isinstance(e[0], NodeOpPar) and e[1] == level:
                return i

def index_cl_par(tree, level, ix_op):
    for i in range(ix_op + 1, len(tree)):
        e = tree[i]
        if isinstance(e, tuple):
            if isinstance(e[0], NodeClPar) and e[1] == level:
                return i

def find_level(tree, level):
    ix_op = index_op_par(tree, level)
    if ix_op is None:
        return None, None
    ix_cl = index_cl_par(tree, level, ix_op)
    if ix_cl is None:
        # unbalanced parenthesis
        return None, None
    return ix_op, ix_cl + 1

def group_tail(tokens):
    i = 0
    if len(tokens) < 1:
        return -1
    if tokens[i][0] == Tok.not_:
        i += 1
    if i > len(tokens):
        return -1
    if tokens[i][0] in tok_pred:
        i += 2
    if i > len(tokens):
        return -1
    if i == 0:
        return -1
    return i

def normalize(nodes):
    res = []
    for node in nodes:
        if isinstance(node, list) and len(node) == 1:
            res.append(node[0])
        else:
            res.append(node)
    return res

class NodeGroupAnd:
    def __init__(self, nodes):
        self._children = normalize(nodes)

    def eval(self, name, path, is_dir, cache):
        for child in self._children:
            res = child.eval(name, path, is_dir, cache)
            if res in [None, False]:
                return res
        # all True
        return True

    def __repr__(self):
        return "NodeGroupAnd({})".format(repr(self._children))

class NodeGroupOr:
    def __init__(self, nodes):
        self._children = normalize(nodes)
    
    def eval(self, name, path, is_dir, cache):
        for child in self._children:
            res = child.eval(name, path, is_dir, cache)
            if res in [None, True]:
                return res
        # all False
        return False

    def __repr__(self):
        return "NodeGroupOr({})".format(repr(self._children))

class NodeGroupNot:
    def __init__(self, children):
        if len(children) != 1:
            raise ValueError("NodeGroupNot unexpected child count {}".format(repr(children)))
        self._child = children[0]

    def eval(self, name, path, is_dir, cache):
        res = self._child.eval(name, path, is_dir, cache)
        if res is None:
            return None
        return not res

    def __repr__(self):
        return "NodeGroupNot({})".format(repr(self._child))

def strip_level(tokens):
    return [e[0] if isinstance(e, tuple) else e for e in tokens]

def pred_token_index(tokens):
    for i, token in enumerate(tokens):
        if not isinstance(token, T):
            continue
        if token.type in tok_pred:
            return i
    return -1

# todo single instance (NodeOr in self._children)

class NodeOpPar:
    def __repr__(self) -> str:
        return '('

class NodeClPar:
    def __repr__(self) -> str:
        return ')'

class NodeNot:
    def __repr__(self) -> str:
        return 'Not'

class NodeAnd:
    def __repr__(self) -> str:
        return 'And'

class NodeOr:
    def __repr__(self) -> str:
        return 'Or'

class NodePred:
    def __init__(self, tokens):
        self._tokens = tokens
        type_, arg = self._type_and_arg()

    def __repr__(self) -> str:
        type_, arg = self._type_and_arg()
        type_str = inv_m[type_]
        if len(self._tokens) == 2:
            return 'NodePred({} {})'.format(type_str, arg)
        else:
            return 'NodePred(-not {} {})'.format(type_str, arg)

    def _type_and_arg(self):
        return self._tokens[-2].type, self._tokens[-1].cont

    def eval(self, name, path, is_dir, cache):

        exp = False if self._tokens[0].type == Tok.not_ else True
        type_, arg = self._type_and_arg()
        
        if arg is None:
            return None

        res = {
            Tok.type: pred_type,
            Tok.mmin: pred_mmin,
            Tok.iname: pred_iname,
            Tok.newer: pred_newer,
            Tok.newermt: pred_newermt,
            Tok.newerct: pred_newerct,
            Tok.ctime: pred_ctime,
            Tok.mtime: pred_mtime,
            Tok.size: pred_size,
            Tok.cont: pred_cont,
            Tok.icont: pred_icont,
            Tok.bcont: pred_bcont,
            Tok.path: pred_path,
            Tok.ipath: pred_ipath,
        }[type_](name, path, is_dir, arg, cache)

        if res is None:
            return None
        return res == exp

def index_of_two(items, pred):
    res = []
    for i, item in enumerate(items):
        res.append(pred(item))
        if res[-2:] == [True, True]:
            return i-1

def indexes_of_two_plus(items, pred):
    b = index_of_two(items, pred)
    if b is None:
        return None, None
    for i in range(b+2, len(items)):
        if not pred(items[i]):
            return b, i
    return b, len(items)

def to_children(nodes):
    #print("to_children({})".format(repr(nodes)))
    if isinstance(nodes[-1], NodeClPar):
        if isinstance(nodes[0], NodeNot):
            return [NodeGroupNot(to_children(nodes[2:-1]))]
        else:
            return to_children(nodes[1:-1])
    else:
        if len(nodes) == 1:
            return nodes

        i = 0
        while True:
            i += 1
            if i > 100:
                t = 1
            b, e = indexes_of_two_plus(nodes, lambda node: not isinstance(node, NodeOr))
            if b is None:
                return [NodeGroupOr(without_or(nodes))]
            nodes = nodes[:b] + [NodeGroupAnd(nodes[b:e])] + nodes[e:]
            if len(nodes) == 1:
                return nodes

def without_or(nodes):
    return [node for node in nodes if not isinstance(node, NodeOr)]

def expr_to_pred(expr):

    level = 0
    
    while True:
        i = pred_token_index(expr)
        if i > -1:
            if i > 0 and isinstance(expr[i-1], T) and expr[i-1].type == Tok.not_:
                head = i-1
            else:
                head = i
            expr = expr[:head] + [NodePred(expr[head: i+2])] + expr[i+2:]
            t = 1
            #print(expr)
        else:
            break

    for i, tok in enumerate(expr):
        if isinstance(tok, T):
            expr[i] = {
                Tok.op_par: NodeOpPar,
                Tok.cl_par: NodeClPar,
                Tok.and_: NodeAnd,
                Tok.or_: NodeOr,
                Tok.not_: NodeNot,
            }[tok.type]()

    expr = [e for e in expr if not isinstance(e, NodeAnd)]

    tree = []

    for i, tok in enumerate(expr):
        
        if isinstance(tok, NodeOpPar):
            level += 1
            tree.append((tok, level))
        elif isinstance(tok, NodeClPar):
            tree.append((tok, level))
            level -= 1
        else:
            tree.append((tok, level))

    if len(tree) == 0:
        return tree, lambda name, path, is_dir: True
    
    while (True):
        level = max_level(tree)
        if level < 0:
            children = strip_level(tree)
            tree = to_children(children)
            break
        h, t = find_level(tree, level)

        if h is None:
            children = strip_level(tree)
            tree = to_children(children)
            break
        else:
            if h > 0 and isinstance(tree[h-1][0], NodeNot):
                h = h-1
            children = strip_level(tree[h:t])
            tree = tree[:h] + [to_children(children)] + tree[t:]

    if len(tree) != 1:
        raise ValueError("Unexpected tree size {}".format(repr(tree)))
    
    tree = tree[0]

    #print(tree)

    cache = Cache()

    return tree, lambda name, path, is_dir: tree.eval(name, path, is_dir, cache)

def walk(top, topdown=True, onerror=None, followlinks=False, maxdepth=0):
    return _walk(os.fspath(top), topdown, onerror, followlinks, maxdepth)

def _walk(top, topdown, onerror, followlinks, maxdepth):
    dirs = []
    nondirs = []
    walk_dirs = []
    try:
        scandir_it = os.scandir(top)
    except OSError as error:
        if onerror is not None:
            onerror(error)
        return

    with scandir_it:
        while True:
            try:
                try:
                    entry = next(scandir_it)
                except StopIteration:
                    break
            except OSError as error:
                if onerror is not None:
                    onerror(error)
                return

            try:
                is_dir = entry.is_dir()
            except OSError:
                is_dir = False

            if is_dir:
                dirs.append(entry.name)
            else:
                nondirs.append(entry.name)

            if not topdown and is_dir:
                if followlinks:
                    walk_into = True
                else:
                    try:
                        is_symlink = entry.is_symlink()
                    except OSError:
                        is_symlink = False
                    walk_into = not is_symlink

                if walk_into:
                    walk_dirs.append(entry.path)

    maxdepth -= 1
    if maxdepth == 0:
        walk_dirs = []

    if topdown:
        yield top, dirs, nondirs
        if maxdepth == 0:
            return
        islink, join = os.path.islink, os.path.join
        for dirname in dirs:
            new_path = join(top, dirname)
            if followlinks or not islink(new_path):
                yield from _walk(new_path, topdown, onerror, followlinks, maxdepth)
    else:
        for new_path in walk_dirs:
            yield from _walk(new_path, topdown, onerror, followlinks, maxdepth)
        yield top, dirs, nondirs

def print_help():
    print("""usage: pyfind [conditions] [-exec cmd args {} ;] [-delete]

finds files and dirs that satisfy conditions (predicates)

predicates:
  -mtime DAYS          if DAYS is negative: modified within DAYS days, 
                       if positive modified more than DAYS days ago
  -ctime DAYS          same as -mtime, but when modified metadata not content
  -mmin MINUTES        if MINUTES is negative: modified within MINUTES minutes, 
                       if positive modified more than MINUTES minutes ago
  -cmin MINUTES        same as -mmin, but when modified metadata not content
  -newer PATH/TO/FILE  modified later than PATH/TO/FILE
  -newermt DATETIME    modified later than DATETIME
  -newerct DATETIME    same as -newermt but when modified metadata not content
  -name PATTERN        filename matches PATTERN (wildcard)
  -iname PATTERN       same as -name but case insensitive
  -path PATTERN        file path matches PATTERN
  -ipath PATTERN       same as -path but case insensitive
  -cont PATTERN        file contains PATTERN
  -icont PATTERN       same as -cont but case insensitive
  -bcont PATTERN       same as -cont but PATTERN is binary expression
  -type d              is directory
  -type f              is file

predicates can be inverted using -not, can be grouped together in boolean expressions 
using -or and -and and parenthesis

examples:
  pyfind -iname *.py -mmin -10
  pyfind -iname *.cpp -or -iname *.h -not ( -iname moc_* -or -iname ui_* )
  pyfind -iname *.h -exec pygrep -H class {} ;
  pyfind -iname *.o -delete
  pyfind -iname *.py | pyxargs pywc -l
""")

def main():

    args = sys.argv[1:]
    if '-h' in args or '--help' in args:
        print_help()
        return

    expr, paths, action, maxdepth = parse_args()
    tree, pred = expr_to_pred(expr)
    if len(paths) == 0:
        paths.append(".")

    for path in paths:
        for root, dirs, files in walk(path, maxdepth=maxdepth):
            for name in dirs:
                p = os.path.join(root, name)
                if pred(name, p, True):
                    action.exec(path, name, p, True)
            for name in files:
                p = os.path.join(root, name)
                if pred(name, p, False):
                    action.exec(path, name, p, False)

def unquote(s):
    if s[0] == '"' and s[-1] == '"':
        return s[1:-1]
    return s

def test():
    args = [unquote(e) for e in '( -iname "lamp.*" -mmin +1 ) -and ( -not -type d )'.split(" ")]
    expr, paths, action = parse_args(args)
    tree, pred = expr_to_pred(expr)

    path = "D:\\w\\lamp.jpg"
    name = os.path.basename(path)
    is_dir = False

    res = pred(name, path, is_dir)

    t = 1

def test2():
    args = "-type f -and ( ( -not -mtime -1 ) -or ( -mmin -1 ) -or -mmin -2 )".split(" ")
    expr, paths = parse_args(args)
    tree, pred = expr_to_pred(expr)

    path = "D:\\w\\test.bat"
    name = os.path.basename(path)
    is_dir = False

    res = pred(name, path, is_dir)

    t = 1

if __name__ == "__main__":
    main()
    #test2()


import sys
import re
from .shared import eprint

def print_help():
    print("""Print numbers from FIRST to LAST with INCREMENT
seq LAST
seq FIRST LAST
seq FIRST INCREMENT LAST
""")

def map_number(args):
    if all([re.match('^[0-9]+$', item) is not None for item in args]):
        return map(int, args), -1
    prec = max([len(item.split('.')[1]) if '.' in item else -1 for item in args])
    return map(float, args), prec

def print_number(n, prec):
    if prec > -1:
        print(("{:." + str(int(prec)) + "f}").format(n))
    else:
        print(n)

def main():
    args = sys.argv[1:]

    if '-h' in args or '--help' in args:
        print_help()
        return

    first = 1
    inc = 1
    last = 0
    prec = -1
    if len(args) == 3:
        (first, inc, last), prec = map_number(args)
    elif len(args) == 2:
        (first, last), prec = map_number(args)
    elif len(args) == 1:
        (last, ), prec = map_number(args)
    else:
        print_help()
        return
    
    i = first

    if inc == 0:
        eprint('inc must be non-zero')
        exit(1)

    elif inc > 0:
        while i <= last:
            print_number(i, prec)
            i += inc
    else:
        while i >= last:
            print_number(i, prec)
            i += inc

if __name__ == "__main__":
    main()


import os
import argparse
from . import read_stdin_text
from .shared import glob_paths_files, print_utf8
import re

def print_cols(line, col_numbers):
    cols = re.split('\\s+', line)
    cols_ = [cols[i-1] if i-1 < len(cols) else '' for i in col_numbers]
    print_utf8(" ".join(cols_))

def main():
    parser = argparse.ArgumentParser(description='extracts and prints specific columns') # todo separators
    parser.add_argument('path', nargs='*')
    parser.add_argument('-n', type=int, nargs='+', help='column number') # todo range expr
    args = parser.parse_args()
    if len(args.path) == 0:
        for line in read_stdin_text().split('\n'):
            print_cols(line, args.n)
    else:
        for path in glob_paths_files(args.path):
            print_cols(path, args.n)

if __name__ == "__main__":
    main()



import argparse
import sys
from .shared import glob_paths, read_bytes, print_lines, drop_last_empty_line, index_of_int
import re

def _print_lines(lines, n, is_head):
    if is_head:
        print_lines(lines[:n])
    else:
        print_lines(lines[-n:])

def bytes_to_lines(bytes_):
    lines = [line + '\n' for line in bytes_.decode('utf-8').split('\n')]
    drop_last_empty_line(lines)
    return lines

def head_tail(is_head = True):

    args_ = sys.argv[1:]

    while True:
        i = index_of_int(args_)
        #print(i)
        if i is None:
            break
        args_ = args_[:i] + ['-n', str(abs(int(args_[i])))] + args_[i+1:]

    description = 'prints n lines from {} of file'.format('head' if is_head else 'tail')

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-n', type=int, default=10, help='number of lines to print')
    parser.add_argument('path', nargs="*")
    args = parser.parse_args(args_)
    paths = glob_paths(args.path)
    stdin_mode = len(paths) == 0

    if stdin_mode:
        bytes_ = read_bytes([])
        lines = bytes_to_lines(bytes_)
        _print_lines(lines, args.n, is_head)
    else:
        for path in paths:
            bytes_ = read_bytes([path])
            lines = bytes_to_lines(bytes_)
            _print_lines(lines, args.n, is_head)


import subprocess
import time
import sys
from .shared import eprint, run
import textwrap

def print_help():
    print(textwrap.dedent("""\
    usage: pytime program [...args]

    measures execution time of application
    """))

def main():
    t1 = time.time()
    
    args = sys.argv[1:]

    if args[0] in ['-h', '--help']:
        print_help()
        return

    #print(args)
    run(args)
    
    t2 = time.time()
    eprint("{:.3f}s".format(t2 - t1))

if __name__ == "__main__":
    main()


import glob
import sys
import io
import subprocess
import os
import re

NUM_RX = r'([-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?)'

def glob_paths(paths):
    res = []
    for path in paths:
        if glob.has_magic(path):
            res += glob.glob(path)
        else:
            res.append(path)
    return res

def glob_paths(paths):
    return _glob_paths_pred(paths, lambda path: True)

def glob_paths_files(paths):
    return _glob_paths_pred(paths, lambda path: os.path.isfile(path))

def glob_paths_dirs(paths):
    return _glob_paths_pred(paths, lambda path: os.path.isdir(path))

def _glob_paths_pred(paths, pred):
    res = []
    for path in paths:
        if glob.has_magic(path):
            for item in glob.glob(path):
                if pred(item):
                    res.append(item)
        else:
            if pred(path):
                res.append(path)
    return res

def glob_paths_dirs(paths):
    res = []
    for path in paths:
        if glob.has_magic(path):
            for item in glob.glob(path):
                if os.path.isdir(item):
                    res.append(item)
        else:
            res.append(path)
    return res

def drop_last_empty_line(lines):
    if lines[-1].strip() == "":
        lines.pop()

"""
def print(arg, encoding='utf-8'):
    if isinstance(arg, bytes):
        print_bytes(arg)
    elif isinstance(arg, list):
        print_lines(arg, encoding)
    elif isinstance(arg, str):
        print_lines([arg], encoding)
"""

def print_bytes(bytes_):
    sys.stdout.buffer.write(bytes_)

def print_lines(lines, encoding='utf-8'):
    for line in lines:
        #sys.stdout.write(line)
        print_bytes(line.encode(encoding))

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def read_bytes(paths):
    if isinstance(paths, list):
        return _read_bytes_many(paths)
    return _read_bytes_one(paths)

def _read_bytes_one(path):
    if path is None:
        return sys.stdin.buffer.read()
    with open(path, 'rb') as f:
        return f.read()

def _read_bytes_many(paths):
    stdin_mode = len(paths) == 0
    if stdin_mode:
        return read_bytes(None)
    else:
        return b''.join([read_bytes(path) for path in paths])

def split_to_lines(text):
    lines = text.split('\n')
    line = lines[-1]
    lines = [line + '\n' for line in lines]
    lines[-1] = line
    return lines

def read_lines(paths, encoding='utf-8', drop_last_empty_line_ = False):
    if isinstance(paths, list):
        return _read_lines_many(paths, encoding, drop_last_empty_line_)
    return _read_lines_one(paths, encoding, drop_last_empty_line_)

def line_reader(paths, from_stdin, encoding='utf-8', drop_last_empty_line_ = False):
    if from_stdin:
        for i, line in enumerate(_read_lines_one(None, encoding, drop_last_empty_line_)):
            yield i, line, '-'
    else:
        for path in paths:
            for i, line in enumerate(_read_lines_one(path, encoding, drop_last_empty_line_)):
                yield i, line, path

def _read_lines_many(paths, encoding='utf-8', drop_last_empty_line_ = False):
    stdin_mode = len(paths) == 0
    if stdin_mode:
        return _read_lines_one(None, encoding, drop_last_empty_line_)
    lines = []
    for path in paths:
        lines += _read_lines_one(path, encoding, drop_last_empty_line_)
    return lines

def _read_lines_one(path, encoding='utf-8', drop_last_empty_line_ = False):
    bytes_ = read_bytes(path)
    lines = split_to_lines(bytes_.decode(encoding))
    if drop_last_empty_line_ and lines[-1] in ['\r', '']:
        lines.pop()
    return lines

def read_lines_(paths, drop_last_empty_line_ = False):
    lines = []

    stdin_mode = len(paths) == 0

    if stdin_mode:
        lines = list(sys.stdin)
        if drop_last_empty_line_:
            drop_last_empty_line(lines)
    else:
        for path in paths:
            with open(path, 'r', encoding='utf-8') as f:
                lines_ = f.readlines()
            if drop_last_empty_line_:
                drop_last_empty_line(lines_)
            lines += lines_

    return lines

def run(args):
    shell = (args[0] in ['type', 'echo', 'copy']) or '|' in args
    subprocess.run(args, shell=shell)


def index_of_int(args):
    for i, arg in enumerate(args):
        if re.match('^-[0-9]+$', arg):
            return i

def parse_args(short, long, short_val, long_val, args):
    opts = {k: False for k in short + long}
    for n in short_val + long_val:
        opts[n] = None
    while len(args) > 0:
        arg = args[0]
        if arg.startswith('--') and arg[2:] in long:
            opts[arg[2:]] = True
        elif arg.startswith('--') and arg[2:] in long_val:
            opts[arg[2:]] = args[1]
            args.pop(0)
        elif arg.startswith('-') and arg[1:] in short_val:
            opts[arg[1:]] = args[1]
            args.pop(0)
        elif arg.startswith('-') and all([c in short for c in arg[1:]]):
            for c in arg[1:]:
                opts[c] = True
        else:
            return opts, args
        args.pop(0)
    return opts, args

def print_utf8(s, end=b'\n'):
    if not isinstance(end, bytes):
        end = end.encode('utf-8')
    sys.stdout.buffer.write(s.encode('utf-8') + end)



import enum
import fileinput
import argparse
import sys
import re
import math
import glob
from types import prepare_class
from .shared import glob_paths, read_bytes

def main():

    description='calculates number or lines words, chars and bytes in files'
    epilog='examples:\n  wc -l text.txt'

    parser = argparse.ArgumentParser(epilog=epilog, description=description, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-l', action='store_true', help='print the newline counts')
    parser.add_argument('-w', action='store_true', help='print the word counts')
    parser.add_argument('-m', action='store_true', help='print the character counts')
    parser.add_argument('-c', action='store_true', help='print the byte counts')
    parser.add_argument('path', nargs="*")

    args = parser.parse_args()

    def count_lines(data, path, res):
        text = ''
        try:
            text = data.decode('utf-8')
        except Exception as e:
            print(e)
        c = len(data)
        m = len(text)
        w = len(re.split('\\s+',text.strip()))
        l = len(text.split('\n'))
        res.append((l, w, m, c, path))

    res = []
    stdin_mode = len(args.path) == 0

    paths = glob_paths(args.path)

    if stdin_mode:
        data = sys.stdin.buffer.read()
        count_lines(data, '', res)
    else:
        for path in paths:
            count_lines(read_bytes(path), path, res)

    lt = sum([e[0] for e in res])
    wt = sum([e[1] for e in res])
    mt = sum([e[2] for e in res])
    ct = sum([e[3] for e in res])

    lw = len(str(lt))
    ww = len(str(wt))
    mw = len(str(mt))
    cw = len(str(ct))

    fmt = " ".join(['{:' + str(lw) + 'd}' if args.l else ''] +
        ['{:' + str(ww) + 'd}' if args.w else ''] +
        ['{:' + str(mw) + 'd}' if args.m else ''] +
        ['{:' + str(cw) + 'd}' if args.c else ''] +
        ['{}'])

    #print(fmt)

    if len(res) > 1:
        res.append((lt,wt,mt,ct,'total'))

    for item in res:
        cols = [v for v, en in zip(item, [args.l, args.w, args.m, args.c, True]) if en]
        print(fmt.format(*cols))

if __name__ == "__main__":
    main()


import argparse
import os

from collections import defaultdict
from itertools import count, zip_longest
import re
from .shared import glob_paths

# todo cleanup unused code
# todo human readible
# todo summary
# todo --files

def int_ceil(num, den):
    return (num + den - 1) // den

def is_path_in(path1, path2):
    if len(path1) < len(path2):
        return False
    for n1, n2 in zip(path1, path2):
        if n1 != n2:
            return False
    return True

def count_equal(it1, it2):
    for i, v1, v2 in zip(count(), it1, it2):
        if v1 != v2:
            return i-1
    return i-1

def path_split_(path):
    path_ = re.split('[/\\\\]', path)
    if re.match('^[a-z]:$',path_[0], re.IGNORECASE):
        path_[0] = path_[0] + '\\'
    return path_

def path_split(path):
    res = []
    rx = re.compile('^([a-z]:\\\\|\\/)$', re.IGNORECASE)
    while True:
        head, tail = os.path.split(path)
        res.append(tail)
        path = head
        if head == '':
            return list(reversed(res))
        if rx.match(head):
            res.append(head)
            return list(reversed(res))

def without_none(items):
    return [v for v in items if v is not None]

def path_diff(path1, path2):
    
    comm = []
    res1 = []
    res2 = []

    is_comm = True
    for n1, n2 in zip_longest(path_split(path1), path_split(path2), fillvalue=None):
        if is_comm:
            if None in [n1, n2]:
                is_comm = False
            else:
                is_comm = n1 == n2
        if is_comm:
            comm.append(n1)
        else:
            res1.append(n1)
            res2.append(n2)
    
    return comm, without_none(res1), without_none(res2)

class Printer:
    def print(self, size, path):
        print("{:10d} {}".format(int_ceil(size, 1024), path))
        pass

class DirStat:
    def __init__(self, printer):
        self._flushed = set()
        self._stat = defaultdict(int)
        self._printer = printer
        self._prev = None

        self._queue = []

    def add_file(self, root, f):
        path = os.path.join(root, f)
        size = os.path.getsize(path)
        path_ = path_split(path)
        for n in range(1, len(path_)):
            self._stat[os.path.join(*path_[:n])] += size
        #self._printer.print(size, path)

    def add_dir(self, root, d):
        path = os.path.join(root, d)
        self._stat[path] = 0

    def begin(self, path):
        self._root = path

    def end(self):
        self.add_root(os.path.dirname(self._root))
        self._queue = []

    def add_root(self, root):

        def pop_queue():
            for i, path in enumerate(self._queue):
                if not root.startswith(path):
                    return self._queue.pop(i)
        
        while True:
            path = pop_queue()
            if path is None:
                break
            #print("flush {}".format(path))
            self._printer.print(self._stat[path], path)
        
        self._queue = [root] + self._queue

    def add_root_(self, root):
        if self._prev is None:
            self._prev = root
            return

        #print(root)

        #print(*path_diff(self._prev, root))
        comm, prev, root_ = path_diff(self._prev, root)

        if len(prev) > 1:
            """
            for i in range(1,len(prev)+1):
                path_ = comm + prev[:i]
                #print(path_)
                print("flush " + os.path.join(*path_))
            """
            path_ = os.path.join(*(comm + [prev[0]]))
            for path, size in self._stat.items():
                if path in self._flushed:
                    continue
                if path.startswith(path_):
                    #print("flush " + path)
                    self._printer.print(size, path)
                self._flushed.add(path)

        self._prev = root

        #print('add_root', root)
        """
        root_ = os.path.split(root)
        for root2, size in self._stat.items():
            if root2 == self._root:
                continue
            if root2 in self._flushed:
                continue
            root2_ = os.path.split(root2)
            if not is_path_in(root2_, root_):
                #print('{} not in {}'.format(root2, root))
                self._flushed.add(root2)
                self._printer.print(size, root2)
        """

def main():
    parser = argparse.ArgumentParser(add_help=False, description='prints directories sizes')
    parser.add_argument('-s',action='store_true', help='print only summary')
    parser.add_argument('-h',action='store_true', help='print in human readible units')
    parser.add_argument('path', nargs='*', help='path to calculate')
    parser.add_argument('--help', action='help')

    args = parser.parse_args()

    paths = glob_paths(args.path)
    if len(paths) == 0:
        paths.append(os.getcwd())

    #print(paths); exit(0)

    printer = Printer()

    stat = DirStat(printer)

    for path in paths:
        #print('os.walk(path)', path)
        stat.begin(path)
        for root, dirs, files in os.walk(path):
            stat.add_root(root)
            for d in dirs:
                stat.add_dir(root, d)
            for f in files:
                stat.add_file(root, f)
        stat.end()

if __name__ == "__main__":
    main()


import argparse
import sys
from .shared import eprint, glob_paths, read_lines, print_lines
import random
import re
from functools import cmp_to_key

NUM_RX = r'([-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?)'

def num_value(s):
    m = re.search(NUM_RX, s)
    if m:
        return float(m.group(1))

def cmp_numeric(s1, s2):
    if s1 == s2:
        return 0
    v1 = num_value(s1)
    v2 = num_value(s2)

    #print(v1, v2)

    if v1 is None and v2 is None:
        return -1 if s1 < s2 else 1
    if v1 is None:
        return -1
    if v2 is None:
        return 1
    if v1 == v2:
        return -1 if s1 < s2 else 1
    return -1 if v1 < v2 else 1
    
def main():
    parser = argparse.ArgumentParser(description='sorts lines')
    parser.add_argument('--numeric-sort', '-n', action='store_true')
    parser.add_argument('--reverse', '-r', action='store_true')
    parser.add_argument('--random-sort', '-R', action='store_true')
    parser.add_argument('--unique', '-u', action='store_true')
    parser.add_argument('path', nargs="*")

    args = parser.parse_args()
    paths = glob_paths(args.path)
    lines = read_lines(paths)
    
    if args.random_sort:
        lines = [(line, random.random()) for line in lines]
        lines.sort(key=lambda item: item[1])
        lines = [e[0] for e in lines]
    elif args.numeric_sort:
        lines.sort(key=cmp_to_key(cmp_numeric))
    else:
        lines.sort()

    if args.unique:
        res = [lines[0]]
        for line in lines:
            if line != res[-1]:
                res.append(line)
        lines = res

    if args.reverse:
        print_lines(reversed(lines))
    else:
        print_lines(lines)

if __name__ == "__main__":
    main()


import argparse
from datetime import date
import datetime
import dateutil.parser
import glob
import os
import sys
from .shared import glob_paths, eprint

def main():
    parser = argparse.ArgumentParser(description='prints file to stdout')
    parser.add_argument('path', nargs='+')
    parser.add_argument('--text', '-t', action='store_true')

    args = parser.parse_args()

    paths = glob_paths(args.path)

    for path in paths:
        try:
            with open(path, 'rb') as f:
                sys.stdout.buffer.write(f.read())
        except Exception as e:
            eprint(e)

if __name__ == "__main__":
    main()


import os
import argparse
from .shared import glob_paths_dirs
from . import read_stdin_text

def main():
    parser = argparse.ArgumentParser(description='prints file extension statistics')
    parser.add_argument("path", nargs="*", help="paths")
    parser.add_argument('-s', '--short', action='store_true')

    args = parser.parse_args()
    
    stat = dict()
    if len(args.path) == 0:
        # read paths from stdin
        text = read_stdin_text()
        for line in text.split('\n'):
            line = line.rstrip()
            ext = os.path.splitext(line)[1]
            if ext not in stat:
                stat[ext] = 0
            stat[ext] += 1
    else:
        # walk paths
        paths = glob_paths_dirs(args.path)
        for path in paths:
            for root, dirs, files in os.walk(path):
                for f in files:
                    ext = os.path.splitext(f)[1]
                    if ext not in stat:
                        stat[ext] = 0
                    stat[ext] += 1

    total = sum(stat.values())
    stat_ = [(ext, count) for ext, count in stat.items()]
    stat_.sort(key= lambda item: item[1], reverse=True)

    t = 0
    for ext, count in stat_:
        print("{:7d} {}".format(count, ext))
        t += count
        if args.short and t / total > 0.9:
            print("{:7d} {}".format(total-t, "other"))
            break
    print("{:7d} {}".format(total, "total"))

if __name__ == "__main__":
    main()


from .shared import glob_paths_files, line_reader, print_lines
import sys
import argparse
import re
import glob

# todo binary files

def main():
    
    parser = argparse.ArgumentParser(add_help=False, description='prints matching lines')
    parser.add_argument('expr')
    parser.add_argument('-i', '--ignore-case', action='store_true', help='ignore case')
    parser.add_argument('-o', '--only-matching', action='store_true')
    parser.add_argument('-v', '--invert-match', action='store_true', help='select non-matching lines')
    parser.add_argument('-H', '--with-filename', action='store_true', help='print file name')
    parser.add_argument('-h', '--without-filename', action='store_true', help='do not print file name')
    parser.add_argument('-n', '--line-number', action='store_true', help='print line number')
    parser.add_argument('--help', action='help')
    parser.add_argument('path', nargs='*')

    args = parser.parse_args()

    paths = glob_paths_files(args.path)

    if args.with_filename:
        print_filename = True
    elif args.without_filename:
        print_filename = False
    else:
        many_inputs = len(args.path) > 1 or (len(args.path) > 0 and glob.has_magic(args.path[0]))
        print_filename = many_inputs

    flags = re.IGNORECASE if args.ignore_case else 0
    rx = re.compile(args.expr, flags)

    def formatter(path, i, m, line):
        cols = []
        if print_filename:
            cols.append(path)
        if args.line_number:
            cols.append(str(i+1))
        if args.only_matching:
            cols.append(m.group(0))
        else:
            cols.append(line)
        return ':'.join(cols) + '\n'

    def pred_match(line):
        m = rx.search(line)
        return m, m is not None

    def pred_not_match(line):
        m = rx.search(line)
        return m, m is None

    if args.invert_match:
        pred = pred_not_match
    else:
        pred = pred_match

    for i, line, path in line_reader(paths, len(args.path) == 0):
        line = line.rstrip('\n')
        m, matched = pred(line)
        if matched:
            print_lines([formatter(path, i, m, line)])

if __name__ == "__main__":
    main()
    


import re
import os
import sys
from .shared import parse_args

"""
def parse_args(args):
    opts = {
        '-e': False,
        '-n': False,
        '-ne': False,
        '-en': False
    }
    while args[0] in opts.keys():
        opt = args.pop(0)
        opts[opt] = True
    if opts['-ne'] or opts['-en']:
        opts['-n'] = True
        opts['-e'] = True
    return opts, args
"""

def print_help():
    print("""usage: pyecho [-e] [-n] [args...]

prints text to stdout

optional arguments:
-h --help  show this message and exit
-e         decode escape sequences
-n         do not print newline
""")
    
def main():
    """
    def env_vars(m):
        n = m.group(1)
        if n in os.environ:
            return os.environ[n]
        return '%' + n + '%'
    def expand_vars(arg):
        return re.sub("%([^%]*)%", env_vars, arg)
    """

    def unescape(arg):
        return arg.encode('utf-8').decode('unicode_escape')

    #print(sys.argv)

    opts, args = parse_args(['e','n','h'],['help'],[],[],sys.argv[1:])

    if opts['h'] or opts['help']:
        print_help()
        return

    def transform(queue, res):

        if len(queue) == 0:
            return False

        arg = queue.pop(0)

        if opts['e']:
            arg = unescape(arg)
        m = re.search("\\{([0-9]+)..([0-9]+)\\}", arg)
        if m:
            head, tail = arg.split(m.group(0), 1)
            start = int(m.group(1))
            end = int(m.group(2))
            for i in reversed(range(start, end+1)):
                queue.insert(0, "{}{}{}".format(head, i, tail))
        else:
            res.append(arg)
        
        return True
        
    #args = [transform(arg) for arg in args]
    
    queue = args
    res = []

    while(transform(queue, res)):
        pass

    text = " ".join(res)
    sys.stdout.buffer.write(text.encode('utf-8'))
    if not opts['n']:
        sys.stdout.buffer.write(b'\n')

if __name__ == "__main__":
    main()


import sys
from .shared import print_lines, glob_paths, read_lines
import re
import argparse

# todo backreferences
# todo -i

def main():
    pass

def parse_exprs(exprs):
    res = []
    for arg in exprs:
        if not arg.startswith("s"):
            raise ValueError("Unexpected expression {}".format(arg))
        sep = arg[1]
        exp = arg.split(sep)
        flags = ""
        if len(exp) == 3: # allow unterminated regexp
            s, subj, repl = exp
        elif len(exp) == 4:
            s, subj, repl, flags = exp
        else:
            raise ValueError("invalid expr {}".format(exp))
        res.append((subj, repl, flags))
    return res

def main():
    
    one_expr = '-e' not in sys.argv[1:]

    parser = argparse.ArgumentParser(description='replaces text')
    if one_expr:
        parser.add_argument('expr')
    parser.add_argument('-e', action='append', help='expression')
    parser.add_argument('path', nargs='*')
    
    args = parser.parse_args()

    if one_expr:
        exprs = parse_exprs([args.expr])
    else:
        exprs = parse_exprs(args.e)

    #print(args); exit(0)

    paths = glob_paths(args.path)
    lines = read_lines(paths)

    def replace(line):
        for subj, repl, flags in exprs:
            flags_ = re.IGNORECASE if 'i' in flags else 0
            line = re.sub(subj, repl, line, flags=flags_)
        return line

    lines = map(replace, lines)
    print_lines(lines)

if __name__ == "__main__":
    main()        



import sys
import re
import subprocess
from .shared import eprint, run, parse_args
from . import read_stdin_text, chunks
import textwrap

def read_stdin_args():
    text = read_stdin_text()
    args = [line for line in [line.strip() for line in text.split('\n')] if line != '']
    return args

def is_int(s):
    try:
        int(s)
        return True
    except Exception as e:
        pass
    return False

def print_help():
    print(textwrap.dedent("""\
    usage: pyxargs [-n num] command [args...]

    reads arguments from stdin and applies them to command

    optional arguments:
    -h --help  show this message and exit
    -n         split arguments to chunks of n
    """))

def main():
    opts, args1 = parse_args(['h'],['help'],['n'],[], sys.argv[1:])
    if opts['h'] or opts['help']:
        print_help()
        return
    args2 = read_stdin_args()
    if opts['n'] is None:
        run(args1 + args2)
    else:
        for args in chunks(args2, int(opts['n'])):
            run(args1 + args)

if __name__ == "__main__":
    main()


import argparse
from datetime import date
import datetime
import dateutil.parser
import glob
import os
from .shared import eprint

def set_mtime(filename, mtime):
    stat = os.stat(filename)
    atime = stat.st_atime
    os.utime(filename, times=(atime, mtime.timestamp()))

def main():
    parser = argparse.ArgumentParser(description='changes mtime to current time or specified time')
    parser.add_argument('path', nargs='+', help='target path')
    parser.add_argument('-d', help='datetime')

    args = parser.parse_args()

    if args.d:
        mtime = dateutil.parser.parse(args.d)
    else:
        mtime = datetime.datetime.now()

    for path in args.path:
        if glob.has_magic(path):
            paths = glob.glob(path)
        else:
            paths = [path]
        for path_ in paths:
            try:
                if os.path.exists(path_):
                    set_mtime(path_, mtime)
                else:
                    with open(path_, 'w') as f:
                        pass
            except Exception as e:
                eprint(e)

if __name__ == "__main__":
    main()


from setuptools import setup, find_packages

with open('readme.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    packages = find_packages(),
    name = 'mugicli',
    version='0.0.3',
    author="Stanislav Doronin",
    author_email="mugisbrows@gmail.com",
    url='https://github.com/mugiseyebrows/mugi-cli',
    description='Shell utilities resembling coreutils and findutils, small extendable and windows-friendly',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    install_requires = ['python-dateutil'],
    entry_points={
        'console_scripts': [
            'pycat = mugicli.pycat:main',
            'pycol = mugicli.pycol:main',
            'pydos2unix = mugicli.pydos2unix:main',
            'pydu = mugicli.pydu:main',
            'pyecho = mugicli.pyecho:main',
            'pyextstat = mugicli.pyextstat:main',
            'pyfind = mugicli.pyfind:main',
            'pygrep = mugicli.pygrep:main',
            'pyhead = mugicli.pyhead:main',
            'pyls = mugicli.pyls:main',
            'pymd5sum = mugicli.pymd5sum:main',
            'pymtime = mugicli.pymtime:main',
            'pymtimestat = mugicli.pymtimestat:main',
            'pysed = mugicli.pysed:main',
            'pyseq = mugicli.pyseq:main',
            'pysha1sum = mugicli.pysha1sum:main',
            'pysha224sum = mugicli.pysha224sum:main',
            'pysha256sum = mugicli.pysha256sum:main',
            'pysha384sum = mugicli.pysha384sum:main',
            'pysha512sum = mugicli.pysha512sum:main',
            'pysort = mugicli.pysort:main',
            'pystart = mugicli.pystart:main',
            'pytail = mugicli.pytail:main',
            'pytime = mugicli.pytime:main',
            'pytmp = mugicli.pytmp:main',
            'pytouch = mugicli.pytouch:main',
            'pyuniq = mugicli.pyuniq:main',
            'pyunix2dos = mugicli.pyunix2dos:main',
            'pywc = mugicli.pywc:main',
            'pyxargs = mugicli.pyxargs:main',
            'pyxxd = mugicli.pyxxd:main',
            'pyzip = mugicli.pyzip:main',
        ]
    },
)


import argparse
import sys
import re
import json
import tempfile
import os
import shutil
from .shared import print_bytes
import textwrap

# todo clean
# todo env PYTMP_DIR

def pytmp_path():
    return os.path.join(os.path.expanduser('~'), '.pytmp')

def write_pathmap(pathmap):
    with open(pytmp_path(), 'w', encoding='utf-8') as f:
        json.dump(pathmap, f, ensure_ascii=False, indent=1)

def read_pathmap():
    try:
        with open(pytmp_path(), encoding='utf-8') as f:
            return json.load(f)
    except Exception as e:
        return dict()

def main():

    epilog = textwrap.dedent("""\
    examples:
        pyecho -e 3\\n1\\n2 | pytmp -o foo
        pytmp -i foo | pysort
        pytmp -p foo | pyxargs pycat 
    """)

    parser = argparse.ArgumentParser(description='temporary file helper', epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-p', '--print', nargs='+', help='print filenames')
    parser.add_argument('-i', '--input', nargs='+', help='input data from file')
    parser.add_argument('-o', '--output', help='output data to file')
    parser.add_argument('--suffix', default='.tmp')

    args = parser.parse_args()

    pathmap = read_pathmap()

    if args.output:
        fd, path = tempfile.mkstemp(prefix=args.output + '-', suffix=args.suffix)
        with open(path, 'wb') as f:
            f.write(sys.stdin.buffer.read())
        os.close(fd)
        pathmap[args.output] = path
        write_pathmap(pathmap)
    elif args.input is not None:
        for n in args.input:
            path = pathmap[n]
            with open(path, 'rb') as f:
                print_bytes(f.read())

    if args.print is not None:
        for n in args.print:
            path = pathmap[n]
            print(path)

if __name__ == "__main__":
    main()